@@ -2,7 +2,7 @@
  * BIOS Decode
  *
  *   (C) 2000-2002 Alan Cox <alan@redhat.com>
- *   (C) 2002-2004 Jean Delvare <khali@linux-fr.org>
+ *   (C) 2002-2005 Jean Delvare <khali@linux-fr.org>
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -51,20 +51,31 @@
  *    http://www.pc.ibm.com/qtechinfo/MIGR-45120.html
  */
 
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <stdio.h>
-#include <unistd.h>
-#include <fcntl.h>
 #include <string.h>
 #include <stdlib.h>
+#include <unistd.h>
+#include <getopt.h>
 
 #include "version.h"
+#include "config.h"
 #include "types.h"
 #include "util.h"
 
+/* Options are global */
+struct opt
+{
+	const char* devmem;
+	unsigned int flags;
+};
+static struct opt opt;
+
+#define FLAG_VERSION            (1<<0)
+#define FLAG_HELP               (1<<1)
+
 struct bios_entry {
 	const char *anchor;
+	size_t anchor_len; /* computed */
 	off_t low_address;
 	off_t high_address;
 	size_t (*length)(const u8 *);
@@ -392,7 +403,7 @@ static size_t compaq_length(const u8 *p)
 
 static int compaq_decode(const u8 *p, size_t len)
 {
-	size_t i;
+	unsigned int i;
 	(void) len;
 
 	printf("Compaq-specific entries present.\n");
@@ -401,16 +412,17 @@ static int compaq_decode(const u8 *p, size_t len)
 	for(i=0; i<p[4]; i++)
 	{
 		/*
-		 * We do not check for truncated entries, because the length was
-		 * computed from the number of records in compaq_length right above,
-		 * so it can't be wrong.
+		 * We do not check for truncated entries, because the length
+		 * was computed from the number of records in compaq_length
+		 * right above, so it can't be wrong.
 		 */
 		if(p[5+i*10]!='$' || !(p[6+i*10]>='A' && p[6+i*10]<='Z')
 			|| !(p[7+i*10]>='A' && p[7+i*10]<='Z')
 			|| !(p[8+i*10]>='A' && p[8+i*10]<='Z'))
 		{
-			printf("\t Abnormal entry! Please report. [%02X %02X %02X %02X]\n",
-				p[5+i*10], p[6+i*10], p[7+i*10], p[8+i*10]);
+			printf("\t Abnormal entry! Please report. [%02X %02X "
+				"%02X %02X]\n", p[5+i*10], p[6+i*10],
+				p[7+i*10], p[8+i*10]);
 			return 0;
 		}
 	}
@@ -476,90 +488,140 @@ static int vpd_decode(const u8 *p, size_t len)
  */
 
 static struct bios_entry bios_entries[]={
-	{ "_SM_", 0xF0000, 0xFFFFF, smbios_length, smbios_decode },
-	{ "_DMI_", 0xF0000, 0xFFFFF, dmi_length, dmi_decode },
-	{ "_SYSID_", 0xE0000, 0xFFFFF, sysid_length, sysid_decode },
-	{ "$PnP", 0xF0000, 0xFFFFF, pnp_length, pnp_decode },
-	{ "RSD PTR ", 0xE0000, 0xFFFFF, acpi_length, acpi_decode },
-	{ "$SNY", 0xE0000, 0xFFFFF, sony_length, sony_decode },
-	{ "_32_", 0xE0000, 0xFFFFF, bios32_length, bios32_decode },
-	{ "$PIR", 0xF0000, 0xFFFFF, pir_length, pir_decode },
-	{ "32OS", 0xE0000, 0xFFFFF, compaq_length, compaq_decode },
-	{ "\252\125VPD", 0xF0000, 0xFFFFF, vpd_length, vpd_decode },
-	{ NULL, 0, 0, NULL, NULL }
+	{ "_SM_", 0, 0xF0000, 0xFFFFF, smbios_length, smbios_decode },
+	{ "_DMI_", 0, 0xF0000, 0xFFFFF, dmi_length, dmi_decode },
+	{ "_SYSID_", 0, 0xE0000, 0xFFFFF, sysid_length, sysid_decode },
+	{ "$PnP", 0, 0xF0000, 0xFFFFF, pnp_length, pnp_decode },
+	{ "RSD PTR ", 0, 0xE0000, 0xFFFFF, acpi_length, acpi_decode },
+	{ "$SNY", 0, 0xE0000, 0xFFFFF, sony_length, sony_decode },
+	{ "_32_", 0, 0xE0000, 0xFFFFF, bios32_length, bios32_decode },
+	{ "$PIR", 0, 0xF0000, 0xFFFFF, pir_length, pir_decode },
+	{ "32OS", 0, 0xE0000, 0xFFFFF, compaq_length, compaq_decode },
+	{ "\252\125VPD", 0, 0xF0000, 0xFFFFF, vpd_length, vpd_decode },
+	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
-int main(int argc, const char *argv[])
+/* Believe it or not, this is significantly faster than memcmp and strncmp */
+static inline int anchor_match(const struct bios_entry *entry, const char *p)
 {
-	u8 buf[16];
-	int fd;
-	off_t fp=0xE0000;
-	const char *devmem="/dev/mem";
+	size_t i;
+	
+	for(i=0; i<entry->anchor_len; i++)
+		if(entry->anchor[i]!=p[i])
+			return 0;
+
+	return 1;
+}
+
+/* Return -1 on error, 0 on success */
+static int parse_command_line(int argc, char * const argv[])
+{
+	int option;
+	const char *optstring = "d:hV";
+	struct option longopts[]={
+		{ "dev-mem", required_argument, NULL, 'd' },
+		{ "help", no_argument, NULL, 'h' },
+		{ "version", no_argument, NULL, 'V' },
+		{ 0, 0, 0, 0 }
+	};
+
+	while((option=getopt_long(argc, argv, optstring, longopts, NULL))!=-1)
+		switch(option)
+		{
+			case 'd':
+				opt.devmem=optarg;
+				break;
+			case 'h':
+				opt.flags|=FLAG_HELP;
+				break;
+			case 'V':
+				opt.flags|=FLAG_VERSION;
+				break;
+			case ':':
+			case '?':
+				return -1;
+		}
+
+	return 0;
+}
+
+static void print_help(void)
+{
+	static const char *help=
+		"Usage: biosdecode [OPTIONS]\n"
+		"Options are:\n"
+		" -d, --dev-mem FILE     Read memory from device FILE (default: " DEFAULT_MEM_DEV ")\n"
+		" -h, --help             Display this help text and exit\n"
+		" -V, --version          Display the version and exit\n";
+	
+	printf("%s", help);
+}
+
+int main(int argc, char * const argv[])
+{
+	u8 *buf;
+	off_t fp;
+	int i;
 	
 	if(sizeof(u8)!=1 || sizeof(u16)!=2 || sizeof(u32)!=4)
 	{
-		fprintf(stderr,"%s: compiler incompatibility\n", argv[0]);
+		fprintf(stderr, "%s: compiler incompatibility\n", argv[0]);
 		exit(255);
 	}
 	
-	if(argc>=2)
-		devmem=argv[1];
-	if((fd=open(devmem, O_RDONLY))==-1 || lseek(fd, fp, SEEK_SET)==-1)
+	/* Set default option values */
+	opt.devmem=DEFAULT_MEM_DEV;
+	opt.flags=0;
+
+	if(parse_command_line(argc, argv)<0)
+		exit(2);
+
+	if(opt.flags & FLAG_HELP)
 	{
-		perror(devmem);
-		exit(1);
+		print_help();
+		return 0;
+	}
+
+	if(opt.flags & FLAG_VERSION)
+	{
+		printf("%s\n", VERSION);
+		return 0;
 	}
 	
 	printf("# biosdecode %s\n", VERSION);
-	while(fp<0xFFFFF)
+
+	if((buf=mem_chunk(0xE0000, 0x20000, opt.devmem))==NULL)
+		exit(1);
+
+	/* Compute anchor lengths once and for all */
+	for(i=0; bios_entries[i].anchor!=NULL; i++)
+		bios_entries[i].anchor_len = strlen(bios_entries[i].anchor);
+
+	for(fp=0xE0000; fp<=0xFFFF0; fp+=16)
 	{
-		int i;
-		
-		if(myread(fd, buf, 16, devmem)==-1)
-			exit(1);
+		u8 *p=buf+fp-0xE0000;
 		
 		for(i=0; bios_entries[i].anchor!=NULL; i++)
 		{
-			if(strncmp((char *)buf, bios_entries[i].anchor, strlen(bios_entries[i].anchor))==0
+			if(anchor_match(&bios_entries[i], (char *)p)
 			 && fp>=bios_entries[i].low_address
 			 && fp<bios_entries[i].high_address)
 			{
-				off_t len=bios_entries[i].length(buf);
-				u8 *p;
+				off_t len=bios_entries[i].length(p);
 				
 				if(fp+len-1<=bios_entries[i].high_address)
 				{
-					if((p=malloc(len))==NULL)
-					{
-						perror("malloc");
-						exit(1);
-					}
-
-					memcpy(p, buf, (len>16?16:len));
-					if(len>16)
-					{
-						/* buffer completion */
-						if(myread(fd, p+16, len-16, devmem)==-1)
-						{
-							free(p);
-							exit(1);
-						}
-					}
 					if(bios_entries[i].decode(p, len))
+					{
 						fp+=(((len-1)>>4)<<4);
-					lseek(fd, fp+16, SEEK_SET);
-					free(p);
+						break;
+					}
 				}
 			}
 		}
-		fp+=16;
-	}
-	
-	if(close(fd)==-1)
-	{
-		perror(devmem);
-		exit(1);
 	}
+
+	free(buf);	
 	
 	return 0;
 }
