@@ -2,7 +2,7 @@
  * DMI Decode
  *
  *   (C) 2000-2002 Alan Cox <alan@redhat.com>
- *   (C) 2002-2004 Jean Delvare <khali@linux-fr.org>
+ *   (C) 2002-2005 Jean Delvare <khali@linux-fr.org>
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -33,10 +33,10 @@
  * information does not come from the above mentioned specification.
  *
  * Additional references:
- *  - Intel AP-485 revision 25
+ *  - Intel AP-485 revision 27
  *    "Intel Processor Identification and the CPUID Instruction"
  *    http://developer.intel.com/design/xeon/applnots/241618.htm
- *  - DMTF Master MIF version 030621
+ *  - DMTF Master MIF version 040707
  *    "DMTF approved standard groups"
  *    http://www.dmtf.org/standards/dmi
  *  - IPMI 1.5 revision 1.1
@@ -48,18 +48,31 @@
  */
 
 #include <stdio.h>
-#include <unistd.h>
 #include <string.h>
 #include <stdlib.h>
+#include <unistd.h>
+#include <getopt.h>
 
 #ifdef __ia64__
 #define USE_EFI
 #endif /* __ia64__ */
 
 #include "version.h"
+#include "config.h"
 #include "types.h"
 #include "util.h"
 
+/* Options are global */
+struct opt
+{
+	const char* devmem;
+	unsigned int flags;
+};
+static struct opt opt;
+
+#define FLAG_VERSION            (1<<0)
+#define FLAG_HELP               (1<<1)
+
 static const char *out_of_spec = "<OUT OF SPEC>";
 static const char *bad_index = "<BAD INDEX>";
 
@@ -674,8 +687,8 @@ static const char *dmi_processor_family(u8 code)
 		NULL, /* 0x77 */
 		"Crusoe TM5000",
 		"Crusoe TM3000",
-		NULL, /* 0x7A */
-		NULL,
+		"Efficeon TM8000",
+		NULL, /* 0x7B */
 		NULL,
 		NULL,
 		NULL,
@@ -685,8 +698,8 @@ static const char *dmi_processor_family(u8 code)
 		"Itanium",
 		"Athlon 64",
 		"Opteron",
-		NULL, /* 0x85 */
-		NULL,
+		"Sempron",
+		NULL, /* 0x86 */
 		NULL,
 		NULL,
 		NULL,
@@ -818,7 +831,7 @@ static const char *dmi_processor_family(u8 code)
 
 static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *prefix)
 {
-	/* Intel AP-485 revision 23, table 5 */
+	/* Intel AP-485 revision 27, table 5 */
 	static const char *flags[32]={
 		"FPU (Floating-point unit on-chip)", /* 0 */
 		"VME (Virtual mode extension)",
@@ -851,7 +864,7 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 		"HTT (Hyper-threading technology)",
 		"TM (Thermal monitor supported)",
 		NULL, /* 30 */
-		"SBF (Signal break on FERR)" /* 31 */
+		"PBE (Pending break enabled)" /* 31 */
 	};
 	/*
 	 * Extra flags are now returned in the ECX register when one calls
@@ -905,7 +918,7 @@ static void dmi_processor_id(u8 type, u8 *p, const char *version, const char *pr
 	else if((type>=0x18 && type<=0x1D) /* AMD */
 	|| type==0x1F /* AMD */
 	|| (type>=0xB6 && type<=0xB7) /* AMD */
-	|| (type==0x83 || type==0x84)) /* AMD 64-bit */
+	|| (type>=0x83 && type<=0x85)) /* AMD */
 		sig=2;
 	else if(type==0x01 || type==0x02)
 	{
@@ -1028,7 +1041,8 @@ static const char *dmi_processor_upgrade(u8 code)
 		"Socket A (Socket 462)",
 		"Socket 478",
 		"Socket 754",
-		"Socket 940" /* 0x11 */
+		"Socket 940",
+		"Socket 939" /* 0x12 */
 	};
 	
 	if(code>=0x01 && code<=0x11)
@@ -1217,7 +1231,7 @@ static void dmi_memory_module_size(u8 code)
 			break;
 		case 0x7F:
 			printf(" Not Installed");
-			break;
+			return;
 		default:
 			printf(" %u MB", 1<<(code&0x7F));
 	}
@@ -1859,7 +1873,7 @@ static const char *dmi_event_log_descriptor_format(u8 code)
 
 static void dmi_event_log_descriptors(u8 count, u8 len, u8 *p, const char *prefix)
 {
-	/* 3.3.16.1, , 3.3.16.6.2 */
+	/* 3.3.16.1 */
 	int i;
 	
 	for(i=0; i<count; i++)
@@ -3321,7 +3335,7 @@ static void dmi_decode(u8 *data, u16 ver)
 			printf("\t\tSBDS Version: %s\n",
 				dmi_string(h, data[0x0E]));
 			printf("\t\tMaximum Error:");
-			dmi_battery_maximum_error(data[0x0E]);
+			dmi_battery_maximum_error(data[0x0F]);
 			printf("\n");
 			if(h->length<0x1A) break;
 			if(data[0x07]==0)
@@ -3342,9 +3356,11 @@ static void dmi_decode(u8 *data, u16 ver)
 			printf("\tSystem Reset\n");
 			if(h->length<0x0D) break;
 			printf("\t\tStatus: %s\n",
-				data[4]&(1<<0)?"Enabled":"Disabled");
+				data[0x04]&(1<<0)?"Enabled":"Disabled");
 			printf("\t\tWatchdog Timer: %s\n",
-				data[4]&(1<<5)?"Present":"No");
+				data[0x04]&(1<<5)?"Present":"Not Present");
+			if(!(data[0x04]&(1<<5)))
+				break;
 			printf("\t\tBoot Option: %s\n",
 				dmi_system_reset_boot_option((data[0x04]>>1)&0x3));
 			printf("\t\tBoot Option On Limit: %s\n",
@@ -3804,11 +3820,54 @@ static int legacy_decode(u8 *buf, const char *devmem)
 }
 #endif /* USE_EFI */
 
-int main(int argc, const char *argv[])
+/* Return -1 on error, 0 on success */
+static int parse_command_line(int argc, char * const argv[])
+{
+	int option;
+	const char *optstring = "d:hV";
+	struct option longopts[]={
+		{ "dev-mem", required_argument, NULL, 'd' },
+		{ "help", no_argument, NULL, 'h' },
+		{ "version", no_argument, NULL, 'V' },
+		{ 0, 0, 0, 0 }
+	};
+
+	while((option=getopt_long(argc, argv, optstring, longopts, NULL))!=-1)
+		switch(option)
+		{
+			case 'd':
+				opt.devmem=optarg;
+				break;
+			case 'h':
+				opt.flags|=FLAG_HELP;
+				break;
+			case 'V':
+				opt.flags|=FLAG_VERSION;
+				break;
+			case ':':
+			case '?':
+				return -1;
+		}
+
+	return 0;
+}
+
+static void print_help(void)
+{
+	static const char *help=
+		"Usage: dmidecode [OPTIONS]\n"
+		"Options are:\n"
+		" -d, --dev-mem FILE     Read memory from device FILE (default: " DEFAULT_MEM_DEV ")\n"
+		" -h, --help             Display this help text and exit\n"
+		" -V, --version          Display the version and exit\n";
+	
+	printf("%s", help);
+}
+
+int main(int argc, char * const argv[])
 {
 	int found=0;
-	off_t fp;
-	const char *devmem="/dev/mem";
+	size_t fp;
 #ifdef USE_EFI
 	FILE *efi_systab;
 	const char *filename;
@@ -3818,12 +3877,28 @@ int main(int argc, const char *argv[])
 	
 	if(sizeof(u8)!=1 || sizeof(u16)!=2 || sizeof(u32)!=4 || '\0'!=0)
 	{
-		fprintf(stderr,"%s: compiler incompatibility\n", argv[0]);
+		fprintf(stderr, "%s: compiler incompatibility\n", argv[0]);
 		exit(255);
 	}
-	
-	if(argc>=2)
-		devmem=argv[1];
+
+	/* Set default option values */
+	opt.devmem=DEFAULT_MEM_DEV;
+	opt.flags=0;
+
+	if(parse_command_line(argc, argv)<0)
+		exit(2);
+
+	if(opt.flags & FLAG_HELP)
+	{
+		print_help();
+		return 0;
+	}
+
+	if(opt.flags & FLAG_VERSION)
+	{
+		printf("%s\n", VERSION);
+		return 0;
+	}
 	
 	printf("# dmidecode %s\n", VERSION);
 	
@@ -3845,7 +3920,7 @@ int main(int argc, const char *argv[])
 		*(addr++)='\0';
 		if(strcmp(linebuf, "SMBIOS")==0)
 		{
-			fp=strtol(addr, NULL, 0);
+			fp=strtoul(addr, NULL, 0);
 			printf("# SMBIOS entry point at 0x%08lx\n", fp);
 		}
 	}
@@ -3857,28 +3932,28 @@ int main(int argc, const char *argv[])
 		exit(1);
 	}
 
-	if((buf=mem_chunk(fp, 0x20, devmem))==NULL)
+	if((buf=mem_chunk(fp, 0x20, opt.devmem))==NULL)
 		exit(1);
 	
-	if(smbios_decode(buf, devmem))
+	if(smbios_decode(buf, opt.devmem))
 		found++;
 	
 	free(buf);
 #else /* USE_EFI */
-	if((buf=mem_chunk(0xF0000, 0x10000, devmem))==NULL)
+	if((buf=mem_chunk(0xF0000, 0x10000, opt.devmem))==NULL)
 		exit(1);
 	
 	for(fp=0; fp<=0xFFF0; fp+=16)
 	{
 		if(memcmp(buf+fp, "_SM_", 4)==0 && fp<=0xFFE0)
 		{
-			if(smbios_decode(buf+fp, devmem))
+			if(smbios_decode(buf+fp, opt.devmem))
 				found++;
 			fp+=16;
 		}
 		else if(memcmp(buf+fp, "_DMI_", 5)==0)
 		{
-			if (legacy_decode(buf+fp, devmem))
+			if (legacy_decode(buf+fp, opt.devmem))
 				found++;
 		}
 	}
