@@ -1,7 +1,7 @@
 /*
  * Compaq Ownership Tag
  *
- *   (C) 2003 Jean Delvare <khali@linux-fr.org>
+ *   (C) 2003-2005 Jean Delvare <khali@linux-fr.org>
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -29,40 +29,40 @@
  *    http://h18000.www1.hp.com/support/techpubs/technical_reference_guides/113a1097.html
  */
 
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <unistd.h>
 #include <string.h>
+#include <unistd.h>
+#include <getopt.h>
 
+#include "version.h"
+#include "config.h"
 #include "types.h"
 #include "util.h"
 
-static void ownership(int fd, u32 base, const char *pname, const char *devmem)
+/* Options are global */
+struct opt
+{
+	const char* devmem;
+	unsigned int flags;
+};
+static struct opt opt;
+
+#define FLAG_VERSION            (1<<0)
+#define FLAG_HELP               (1<<1)
+
+static void ownership(u32 base, const char *pname, const char *devmem)
 {
 	u8 *buf;
 	int i;
-	
-	if((buf=malloc(0x51))==NULL)
+
+	/* read the ownership tag */
+	if((buf=mem_chunk(base, 0x51, devmem))==NULL)
 	{
 		perror(pname);
 		return;
 	}
 	
-	/* read the ownership tag */
-	if(lseek(fd, (off_t)base, SEEK_SET)==-1)
-	{
-		perror(devmem);
-		return;
-	}
-	if(myread(fd, buf, 0x50, devmem)==-1)
-	{
-		free(buf);
-		exit(1);
-	}
-	
 	/* chop the trailing garbage */
 	i=0x4f;
 	while(i>=0 && (buf[i]==0x20 || buf[i]==0x00))
@@ -108,76 +108,106 @@ static u32 decode(const u8 *p)
 	return 0;
 }
 
-int main(int argc, const char *argv[])
+/* Return -1 on error, 0 on success */
+static int parse_command_line(int argc, char * const argv[])
 {
-	u8 buf[16];
-	int fd;
-	off_t fp=0xE0000;
-	const char *devmem="/dev/mem";
+	int option;
+	const char *optstring = "d:hV";
+	struct option longopts[]={
+		{ "dev-mem", required_argument, NULL, 'd' },
+		{ "help", no_argument, NULL, 'h' },
+		{ "version", no_argument, NULL, 'V' },
+		{ 0, 0, 0, 0 }
+	};
+
+	while((option=getopt_long(argc, argv, optstring, longopts, NULL))!=-1)
+		switch(option)
+		{
+			case 'd':
+				opt.devmem=optarg;
+				break;
+			case 'h':
+				opt.flags|=FLAG_HELP;
+				break;
+			case 'V':
+				opt.flags|=FLAG_VERSION;
+				break;
+			case ':':
+			case '?':
+				return -1;
+		}
+
+	return 0;
+}
+
+static void print_help(void)
+{
+	static const char *help=
+		"Usage: ownership [OPTIONS]\n"
+		"Options are:\n"
+		" -d, --dev-mem FILE     Read memory from device FILE (default: " DEFAULT_MEM_DEV ")\n"
+		" -h, --help             Display this help text and exit\n"
+		" -V, --version          Display the version and exit\n";
+	
+	printf("%s", help);
+}
+
+int main(int argc, char * const argv[])
+{
+	u8 *buf;
+	off_t fp;
 	int ok=0;
 	
 	if(sizeof(u8)!=1 || sizeof(u32)!=4)
 	{
-		fprintf(stderr,"%s: compiler incompatibility\n", argv[0]);
+		fprintf(stderr, "%s: compiler incompatibility\n", argv[0]);
 		exit(255);
 	}
-	
-	if(argc>=2)
-		devmem=argv[1];
-	if((fd=open(devmem, O_RDONLY))==-1 || lseek(fd, fp, SEEK_SET)==-1)
+
+	/* Set default option values */
+	opt.devmem=DEFAULT_MEM_DEV;
+	opt.flags=0;
+
+	if(parse_command_line(argc, argv)<0)
+		exit(2);
+
+	if(opt.flags & FLAG_HELP)
 	{
-		perror(devmem);
-		exit(1);
+		print_help();
+		return 0;
 	}
 
-	while(!ok && fp<0xFFFFF)
+	if(opt.flags & FLAG_VERSION)
 	{
-		if(myread(fd, buf, 16, devmem)==-1)
-			exit(1);
-		
-		if(memcmp((char *)buf, "32OS", 4)==0)
-		{
-			off_t len=buf[4]*10+5;
-			u8 *p;
+		printf("%s\n", VERSION);
+		return 0;
+	}
+	
+	if((buf=mem_chunk(0xE0000, 0x20000, opt.devmem))==NULL)
+		exit(1);
 
-			if(fp+len-1<=0xFFFFF)
+	for(fp=0; !ok && fp<=0x1FFF0; fp+=16)
+	{
+		u8 *p=buf+fp;
+
+		if(memcmp((char *)p, "32OS", 4)==0)
+		{
+			off_t len=p[4]*10+5;
+
+			if(fp+len-1<=0x1FFFF)
 			{
 				u32 base;
 				
-				if((p=malloc(len))==NULL)
-				{
-					perror("malloc");
-					exit(1);
-				}
-
-				memcpy(p, buf, 16);
-				if(len>16)
-				{
-					/* buffer completion */
-					if(myread(fd, p+16, len-16, devmem)==-1)
-					{
-						free(p);
-						exit(1);
-					}
-				}
 				if((base=decode(p)))
 				{
 					ok=1;
-					ownership(fd, base, argv[0], devmem);
+					ownership(base, argv[0], opt.devmem);
 				}
-				else
-					lseek(fd, fp+16, SEEK_SET);
-				free(p);
 			}
 		}
-		fp+=16;
 	}
 	
-	if(close(fd)==-1)
-	{
-		perror(devmem);
-		exit(1);
-	}
+	free(buf);
 	
 	return 0;
 }
