@@ -1,7 +1,7 @@
 /*
  * IBM Vital Product Data decoder
  *
- *   (C) 2003-2004 Jean Delvare <khali@linux-fr.org>
+ *   (C) 2003-2005 Jean Delvare <khali@linux-fr.org>
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -25,32 +25,46 @@
  *
  * References:
  *  - IBM "Using the BIOS Build ID to identify IBM Thinkpad systems"
- *    Revision "November 17, 2003"
- *    http://www.pc.ibm.com/qtechinfo/MIGR-45120.html
+ *    Revision 2005-02-16
+ *    http://www-307.ibm.com/pc/support/site.wss/MIGR-45120.html
  *
  * Notes:
  *  - Main part of the code is taken directly from biosdecode, with an
  *    additional lookup table for the product name.
  */
 
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <unistd.h>
 #include <string.h>
+#include <unistd.h>
+#include <getopt.h>
 
 #include "version.h"
+#include "config.h"
 #include "types.h"
 #include "util.h"
 
+/* Options are global */
+struct opt
+{
+	const char* devmem;
+	unsigned int flags;
+};
+static struct opt opt;
+
+#define FLAG_VERSION            (1<<0)
+#define FLAG_HELP               (1<<1)
+
 static const char *product_name(const char *id)
 {
+	/* Please keep IDs in alphabetical order */
 	static const char *name[]={
+		"AP", "eServer xSeries 336", /* added 2005-02-01,
+		                                reported by David Rosala,
+		                                confirmed by Scott Denham */
 		"GE", "eServer xSeries 345", /* added 2004-06-11,
 		                                reported by Doug Brenner */
-		"HR", "Thinkpad 650E",
+		"HR", "Thinkpad 560E",
 		"HV", "Thinkpad 760C/CD/L/LD",
 		"HX", "Thinkpad 760E/ED/X/XD/XL or 765/L/D (9385XGA)",
 		"HY", "Thinkpad 760E/EL/ELD (9320SVGA)",
@@ -89,6 +103,9 @@ static const char *product_name(const char *id)
 		"IZ", "Thinkpad X20 or X21", /* updated 2003-11-29 (IBM) */
 		"JP", "eServer xSeries 205", /* added 2004-04-30,
 		                                reported by Bernd Krumboeck */
+		"KP", "eServer xSeries 346", /* added 2005-02-01,
+		                                reported by David Rosala,
+		                                confirmed by Scott Denham */
 		"KQ", "Thinkpad i1200 or i1300",
 		"KR", "Thinkpad i1400 or i1500",
 		"KS", "Thinkpad 240X",
@@ -103,6 +120,8 @@ static const char *product_name(const char *id)
 		"KZ", "Thinkpad T21", /* fixed 2003-11-29 (IBM) */
 		"NV", "PC 300PL", /* added 2004-04-02,
 		                     reported by Shawn Starr */
+		"OP", "Intellistation Z10", /* added 2005-02-17,
+		                               reported by Scott Denham */
 		"PD", "PC 300GL", /* added 2004-04-14,
 		                     reported by Roger Koot */
 		"PI", "Netvista A40/p", /* added 2004-02-23,
@@ -112,10 +131,14 @@ static const char *product_name(const char *id)
 		                     reported by Roger Koot */
 		"PL", "Intellistation M-Pro", /* added 2004-04-15,
 		                                 reported by Mark Syms */
+		"PN", "Intellistation A10", /* added 2005-02-17,
+		                               reported by Scott Denham */
 		"PT", "Netvista A20", /* added 2003-12-28,
 		                         reported by Ramiro Barreiro */
 		"RE", "eServer xSeries 445", /* added 2003-12-17,
 		                                reported by Josef Moellers */
+		"RD", "eServer xSeries 365", /* added 2005-02-01,
+		                                reported by David Rosala */
 		"T2", "eServer xSeries 335", /* added 2004-06-11,
 		                                reported by Doug Brenner */
 		"TT", "eServer xSeries 330", /* added 2003-12-03,
@@ -147,6 +170,8 @@ static const char *product_name(const char *id)
 								reported by Marco Wertejuk */
 		"1S", "Thinkpad R40e", /* added 2003-11-29 (IBM) */
 		"1T", "Thinkpad G40",
+		"1W", "Thinkpad R50e", /* added 2005-02-17 (IBM) */
+		"1Y", "Thinkpad T43/p", /* added 2005-02-17 (IBM) */
 		"20", "Netvista A22p or M41", /* added 2003-10-09,
 		                                 updated 2004-02-23,
 		                                 updated 2004-03-24 */
@@ -154,6 +179,8 @@ static const char *product_name(const char *id)
 		                         reported by Paul Sturm */
 		"2A", "Thinkcentre M50", /* added 2004-03-19,
 		                            reported by Rafael Avila de Espindola */
+		"2C", "Thinkcentre A50", /* added 2005-02-24,
+		                            reported by Tomek Mateja */
 		NULL, "Unknown, please report!"
 	};
 	
@@ -218,71 +245,97 @@ static int decode(const u8 *p)
 	return 1;
 }
 
-int main(int argc, const char *argv[])
+/* Return -1 on error, 0 on success */
+static int parse_command_line(int argc, char * const argv[])
 {
-	u8 buf[16];
-	int fd, found=0;
-	off_t fp=0xF0000;
-	const char *devmem="/dev/mem";
+	int option;
+	const char *optstring = "d:hV";
+	struct option longopts[]={
+		{ "dev-mem", required_argument, NULL, 'd' },
+		{ "help", no_argument, NULL, 'h' },
+		{ "version", no_argument, NULL, 'V' },
+		{ 0, 0, 0, 0 }
+	};
+
+	while((option=getopt_long(argc, argv, optstring, longopts, NULL))!=-1)
+		switch(option)
+		{
+			case 'd':
+				opt.devmem=optarg;
+				break;
+			case 'h':
+				opt.flags|=FLAG_HELP;
+				break;
+			case 'V':
+				opt.flags|=FLAG_VERSION;
+				break;
+			case ':':
+			case '?':
+				return -1;
+		}
+
+	return 0;
+}
+
+static void print_help(void)
+{
+	static const char *help=
+		"Usage: vpddecode [OPTIONS]\n"
+		"Options are:\n"
+		" -d, --dev-mem FILE     Read memory from device FILE (default: " DEFAULT_MEM_DEV ")\n"
+		" -h, --help             Display this help text and exit\n"
+		" -V, --version          Display the version and exit\n";
+	
+	printf("%s", help);
+}
+
+int main(int argc, char * const argv[])
+{
+	u8 *buf;
+	int found=0;
+	off_t fp;
 	
 	if(sizeof(u8)!=1)
 	{
-		fprintf(stderr,"%s: compiler incompatibility\n", argv[0]);
+		fprintf(stderr, "%s: compiler incompatibility\n", argv[0]);
 		exit(255);
 	}
 	
-	if(argc>=2)
-		devmem=argv[1];
-	if((fd=open(devmem, O_RDONLY))==-1 || lseek(fd, fp, SEEK_SET)==-1)
+	/* Set default option values */
+	opt.devmem=DEFAULT_MEM_DEV;
+	opt.flags=0;
+
+	if(parse_command_line(argc, argv)<0)
+		exit(2);
+
+	if(opt.flags & FLAG_HELP)
 	{
-		perror(devmem);
-		exit(1);
+		print_help();
+		return 0;
+	}
+
+	if(opt.flags & FLAG_VERSION)
+	{
+		printf("%s\n", VERSION);
+		return 0;
 	}
 
 	printf("# vpddecode %s\n", VERSION);
-	while(fp<0xFFFFF)
-	{
-		if(myread(fd, buf, 16, devmem)==-1)
-			exit(1);
-		
-		if(memcmp((char *)buf, "\252\125VPD", 5)==0)
-		{
-			off_t len=buf[5];
-			u8 *p;
 
-			if(fp+len-1<=0xFFFFF)
-			{
-				if((p=malloc(len))==NULL)
-				{
-					perror(argv[0]);
-					exit(1);
-				}
-
-				memcpy(p, buf, 16);
-				if(len>16)
-				{
-					/* buffer completion */
-					if(myread(fd, p+16, len-16, devmem)==-1)
-					{
-						free(p);
-						exit(1);
-					}
-					lseek(fd, fp+16, SEEK_SET);
-				}
-
-				if(decode(p))
-					found++;
-				free(p);
-			}
-		}
-		fp+=16;
-	}
-	
-	if(close(fd)==-1)
-	{
-		perror(devmem);
+	if((buf=mem_chunk(0xF0000, 0x10000, opt.devmem))==NULL)
 		exit(1);
+
+	for(fp=0; fp<=0xFFF0; fp+=16)
+	{
+		u8 *p=buf+fp;
+
+		if(memcmp((char *)p, "\252\125VPD", 5)==0
+		&& fp+p[5]-1<=0xFFFF
+		&& decode(p))
+			found++;
 	}
+
+	free(buf);
 	
 	if(!found)
 		printf("# No VPD structure found, sorry.\n");
